package se.chalmers.paragon;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;

/**
 * Represents a clause of a policy
 * 
 * @author Paragon Developers (http://www.cse.chalmers.se/paragon)
 * 
 */
public class Clause {

	/**
	 * The head of the clause
	 */
	private Atom head;

	/**
	 * List of lock pairs that make the body of the clause
	 */
	private Atom[] body;

	/**
	 * Create a new clause. Adds the clause as a relevant clause to the lock in
	 * the head of the clause
	 * 
	 * @param head
	 *            Head of the clause
	 * @param body
	 *            Body of the clause
	 */
	private Clause(Atom head, Atom... body) {
		this.head = head;
		this.body = body;
	}

	/**
	 * Create a new clause.
	 * 
	 * @param head
	 *            Head of the clause
	 * @param body
	 *            Body of the clause
	 * @return The new clause
	 */
	public static Clause newClause(Atom head, Atom... body) {

		// Make the rule safe by binding all unbounded variables in the head
		// with the isActor lock

		HashSet<ActorVariable> bodyVars = new HashSet<ActorVariable>();
		for (Atom b : body)
			bodyVars.addAll(b.getArgs().getVariables());
		HashSet<ActorVariable> unbounded = head.getArgs().getVariables();
		unbounded.removeAll(bodyVars);

		if (unbounded.size() > 0) {
			LinkedList<Atom> _body = new LinkedList<Atom>(Arrays.asList(body));
			for (ActorVariable v : unbounded)
				_body.add(Atom.newAtom(Lock.ISACTOR, v));
			body = _body.toArray(body);
		}

		return new Clause(head, body);
	}

	/**
	 * Deduce for this clause only all possible new heads
	 * 
	 * @param varmap
	 *            Current mapping of variables into concrete actors
	 * @return True if new facts were derived, false otherwise
	 */
	protected boolean deduceOneStep(HashMap<ActorVariable, ConcreteActor> varmap) {
		Atom[] s_body = new Atom[body.length];
		for (int i = 0; i < body.length; i++)
			s_body[i] = Util.substitute(body[i], varmap);
		for (Atom lp : s_body) {
			if (lp.isGround()) {
				if (lp.getLock().isExplicitOrImplicitOpen(
						lp.getArgs().getActorsConcrete()))
					continue;
				else
					return false; // clause can not be applied
			} else {
				HashSet<ActorList> _aps = lp.getLock().getAllOpen(
						lp.getArgs().getActors());
				boolean newFacts = false;
				for (ActorList ap : _aps) {
					newFacts = newFacts
							|| deduceOneStep(Util.extendVarMap(varmap, lp, ap));
				}
				return newFacts; // Recursion over
			}
		}
		// Arriving here means all locks in body are ground and open
		// So we can conclude the head of the clause
		return head.getLock().openImplicit(
				Util.substitute(head, varmap).getArgs().getActorsConcrete());
	}

	/**
	 * Iterates over all locks occurring in the body and make them try and
	 * deduce new facts (in one step)
	 * 
	 * @param locksIterated
	 *            The locks having deduced new facts so far
	 * 
	 * @return True if new facts were derived, false otherwise
	 */
	protected boolean deduceRelevantLocksOneStep(HashSet<Lock> locksIterated) {
		boolean newFacts = false;
		for (Atom lp : body) {
			if (locksIterated.contains(lp.getLock()))
				continue;
			newFacts = newFacts || lp.getLock().deduceOneStep(locksIterated);
		}
		return newFacts;
	}

	/**
	 * Calls to all locks in the body to remove the derived clauses
	 * 
	 * @param locksIterated
	 *            The locks having deduced new facts so far
	 */
	public void cleanUpImplicit(HashSet<Lock> locksIterated) {
		for (Atom lp : body) {
			if (locksIterated.contains(lp.getLock()))
				continue;
			lp.getLock().cleanUpImplicit(locksIterated);
		}
	}

	protected Atom[] getBody() {
		return body;
	}

	protected Atom getHead() {
		return head;
	}

	public String toString() {
		String res = "";
		res += getHead().toString();
		res += " :- ";
		for (Atom b : getBody()) {
			res += b.toString() + ",";
		}
		res += ".";
		return res.replaceAll(",\\.", ".");
	}

}
